const {log, warn, error, debug} = require('zog')
import config from 'configurize'
import _ from 'lodash'
import {Disposable} from 'event-kit'
import dedent from 'dedent'

export default class {

  proxy;

  services = {
    consume: {
      'app-server.service': {'1.0.0': 'consumeAppServerService'}
    }
  };

  consumeAppServerService(service) {
    service.addFirstMiddleware(this.getMiddleware())
    service.addHook('start-listening', (koaApp) => {
      // legacy method was 'app-server:listen'
      koaApp.httpServer.on('upgrade', (req, socket, head) => {
        if (req.url.indexOf('/webpack-dev-server-proxy') == -1) return
        const newUrl = req.url.replace('/webpack-dev-server-proxy', '')
        req.url = newUrl
        debug('Proxying websockets to:', newUrl)
        this.proxy.ws(req, socket, head)
      })
    })
    return new Disposable(() => { })
  }

  getMiddleware() {
    const middleware = [
      this.faviconMiddleware(),
      this.staticAssetsMiddleware(),
      this.appAssetsMiddleware(),
    ]
    return _.flatten(middleware)
  }

  faviconMiddleware() {
    const favicon = require('koa-favicon')
    const convert = require('koa-convert')
    const mw = convert(favicon('public/favicon.ico'))
    mw._name = 'public/favicon.ico'
    return mw
  }

  staticAssetsMiddleware() {
    const serveStatic = require('koa-serve-static')
    const mount = require('koa-mount-hot')
    const mw = mount('/public', serveStatic('public', {
      // NOTE: Must be false, otherwise server blocks.
      // See: https://github.com/expressjs/serve-static#fallthrough
      fallthrough: false,
    }))
    mw._name = 'mount /public'
    return mw
  }

  // Uses `http-proxy` module to forward requests to particular paths to webpack-dev-server.
  // NOTE: `koa-proxy` did not seem to work.
  appAssetsMiddleware() {

    const middleware = []

    // TODO(vjpr): Make this nicer.
    // Get dev server port.
    const {webpackeratorUtils} = require('webpackerator')
    const opts = webpackeratorUtils.parseOpts({}) // TODO(vjpr): Not the best. Need single source of truth for config.
    // TODO(vjpr): Don't use this its super slow!
    //const webpackConfig = webpackeratorUtils.getConfig(opts)
    //const {port: webpackDevServerPort} = webpackConfig.devServer
    //// TODO(vjpr): Verify the dev-server corresponds to this project. There may my multiple running.
    const webpackDevServerPort = opts.devServerPort || 8082
    //

    console.log('Using webpack dev server on port:', webpackDevServerPort)

    const proxy = this.proxy = require('http-proxy').createProxyServer({
      target: {host: '0.0.0.0', port: webpackDevServerPort},
      ws: true,
    })
    proxy.on('error', (e) => {
      if (e.code === 'ECONNREFUSED') {
        const errStr = `
          Could not connect to webpack-dev-server.
          A request was made which is to be forwarded to a webpack-dev-server.
            -> Ensure it is running on port ${webpackDevServerPort} by running 'gulp webpack:dev-server'.
          `
        error(dedent(errStr))
        return
      }
      error('PROXY ERROR:', e)
    })
    const route = require('koa-route')

    const handler = async (ctx, next) => {
      ctx.respond = false
      console.log('proxying', ctx.request.url)
      proxy.web(ctx.req, ctx.res)
    }

    //
    // webpack-dev-server hot reloading
    //

    // NOTE: This handler is for webpack-dev-server hot reloading, not for asset serving.
    const webpackDevServerProxy = async (ctx, next) => {
      ctx.respond = false
      const newUrl = ctx.req.url.replace('/webpack-dev-server-proxy', '')
      ctx.req.url = newUrl
      //debug('Proxying:', ctx.req.url)
      proxy.web(ctx.req, ctx.res)
    }

    // ---

    // TODO(vjpr): Should this be koa-mount-hot?
    function serveStaticMiddleware(mountPath, dir) {
      // NOTE: Nginx will prob be used so this is never hit.
      const serveStatic = require('koa-serve-static')
      const mount = require('koa-mount-hot')
      //const route = require('koa-route')
      const middleware = mount(mountPath, serveStatic(dir, {
      //const middleware = route.get(mountPath, serveStatic(dir, {
        maxAge: 60000,
        etag: true,
      }))
      middleware._name = 'koa-serve-static serveStaticMiddleware'
      return middleware
    }

    let serveAssetsFromWebpackDevServer = config.get('envProps.serveAssetsFromWebpackDevServer')
    if (_.isUndefined(serveAssetsFromWebpackDevServer)) {
      serveAssetsFromWebpackDevServer = true
    }
    console.log('Serving assets from webpack dev server: ' + serveAssetsFromWebpackDevServer)

    if (serveAssetsFromWebpackDevServer) {
      let mw

      middleware.push(serveStaticMiddleware('/build-noproxy', 'build'))
      //route.get('/hot/*', handler),

      mw = route.get('/build/info', async () => {
        console.error('webpack-dev-server/client?=xxx is incorrect')
        return
      })
      mw._name = 'build/info'
      middleware.push(mw)

      mw = route.get('/build/*', handler)
      mw._name = '/build/*'
      middleware.push(mw)

      mw = route.all('/webpack-dev-server-proxy/*', webpackDevServerProxy)
      mw._name = '/webpack-dev-server-proxy/*'
      middleware.push(mw)

    } else {
      middleware.push(serveStaticMiddleware('/build', 'build'))
    }

    //redbirdProxy()

    return middleware

  }

}

function redbirdProxy() {

  // This is an alternative way to reverse proxy to webpack dev server and backend.
  const redbirdProxy = require('redbird')({
    port: 3001,
    bunyan: false, // Disable logging.
  })
  redbirdProxy.register('localhost', 'http://localhost:3000')
  redbirdProxy.register('localhost/build', 'http://localhost:8081/build')

}
